defmodule Verba.Adjective do
  alias Verba.Adjective, as: Adjective

  alias Verba.Decline, as: Decline
  alias Verba.Inflection, as: Inflection

  @typedoc """
  Regular Latin adjectives belong to one of two declension groups.
  """
  @type group() :: :first_second | :third

  @typedoc """
  Regular third declension adjectives can have one, two, or three terminations.
  """
  @type termination() :: :one | :two | :three

  @typedoc """
  Holds the information necessary to properly decline adjectives.

  * `:declension`: Holds the singular nominative form for all genders. These
                   are used to decline the adjective.
  * `:attr`: Holds special attributes for an adjectives.

  # Attributes

  * `:c_stem`: Third declension adjectives can have a consonant stem or an
      i-stem, which changes a way it declines in a few cases. If attr contains
      the key :c_stem, then the adjective is a c-stem, otherwise it's an
      i-stem.

  * `:ius_genitive_ending: Nine first and second declension adjectives have
      distinctive -īus singular genitive and -ī singular dative endings for all
      three genders. Those adjectives are:

      ūllus, ūlla, ūllum
      nūllus, nūlla, nūllum
      uter, utra, utrum
      sōlus, sōla, sōlum
      neuter, neutra, neutrum
      alius, alia, aliud
      ūnus, ūna, ūnum
      tōtus, tōta, tōtum
      alter, altera, alterum

      Adding this key to an adjective's attr map will cause it to decline in
      the manner of these nine adjectives.
  """
  defstruct declension: %{}, attr: %{}

  @opaque adjective() :: %Adjective{}

  # Takes the dictionary form of an adjective and returns a Map containing the
  # nominative forms for all three genders. Also, in the case of third
  # declension adjectives with one termination, the returned Map will also
  # include the genitive form for all three genders as that information is
  # required to both determine the adjective's declension group and provide the
  # adjective's stem.
  @spec create_declension([String.t()]) :: {atom(), map()} | {atom(), String.t()}
  defp create_declension(dict) do
    case group(dict) do
      {:first_second, _} -> {:ok, create_first_second_declension(dict)}
      {:third, :one} -> {:ok, create_third_one_declension(dict)}
      {:third, :two} -> {:ok, create_third_two_declension(dict)}
      {:third, :three} -> {:ok, create_third_three_declension(dict)}
      nil -> {:error, "The dictionary form provided does not match any adjective declension group."}
    end
  end

  @spec create_first_second_declension([String.t()]) :: map()
  defp create_first_second_declension([masculine | [feminine | [neuter | []]]]) do
    %{
      :singular => %{
        :nominative => %{
          :masculine => [masculine],
          :feminine => [feminine],
          :neuter => [neuter],
        },
      },
    }
  end

  @spec create_third_one_declension([String.t()]) :: map()
  defp create_third_one_declension([nominative | [genitive | []]]) do
    %{
      :singular => %{
        :nominative => %{
          :masculine => [nominative],
          :feminine => [nominative],
          :neuter => [nominative],
        },
        :genitive => %{
          :masculine => [genitive],
          :feminine => [genitive],
          :neuter => [genitive],
        },
      },
    }
  end

  @spec create_third_two_declension([String.t()]) :: map()
  defp create_third_two_declension([masculine | [neuter | []]]) do
    %{
      :singular => %{
        :nominative => %{
          :masculine => [masculine],
          :feminine => [masculine],
          :neuter => [neuter],
        },
      },
    }
  end

  @spec create_third_three_declension([String.t()]) :: map()
  defp create_third_three_declension([masculine | [feminine | [neuter | []]]]) do
    %{
      :singular => %{
        :nominative => %{
          :masculine => [masculine],
          :feminine => [feminine],
          :neuter => [neuter],
        },
      },
    }
  end

  @doc """
  Creates a regular first and second declension Latin adjective.

  The dictionary form of a regular first and second declension Latin adjective
  consists of the singular nominative form for all three genders. For example,
  for the first and second declension adjecive altus (high, tall) the
  dictionary form would be altus, alta, altum.

  This dictionary form of regular third declension adjectives depends on
  whether the adjective has one, two, or three terminations. Third declension
  adjectives with one termination require the singular nominative and genitive
  forms. For example, for the adjective atrōx (terrible) the dictionary form
  would be atrōx, atrōcis.

  Third declension adjectives with two terminations consists of two forms: the
  masculine and feminine singular nominative and the neuter singular
  nominative. For example, for the adjective agilis (swift) the dictionary form
  would be agilis, agile.

  Third declension adjectives with three terminations consist of three forms:
  the masculine, feminine, and neuter singular nominatives. For example, for
  the adjective alacer (lively) the dictionary form would be alacer, alacris,
  alacre.

  This function takes the dictionary form of a regular Latin adjective.
  Depending on the form passed in, it will return a first and second declension
  adjective or a third declension adjective with one, two, or three
  three terminations.

  ## Warning

  Third declension adjectives can be either consonant-stem or i-stem. Third
  declension adjectives created with this function will decline as i-stems.
  When creating a third declension consonant-stem adjective, use
  `regular_c_stem` instead.

  ## Example

      iex> {:ok, adjective} = Verba.Adjective.regular(["altus", "alta", "altum"])
      {:ok,
        %Verba.Adjective{
          attr: %{},
          declension: %{
            singular: %{
              nominative: %{
                feminine: ["alta"],
                masculine: ["altus"],
                neuter: ["altum"]
              }
            }
          }
        }}
      iex> Verba.Adjective.decline(adjective, :singular, :dative, :masculine)
      ["altō"]

  """
  @spec regular([String.t()]) :: {atom(), adjective()} | {atom(), String.t()}
  def regular(dict = [_|_]) do
    case create_declension(Decline.normalize_declension(dict)) do # Since regular_c_stem/1 and irregular_ius_genitive/1 call this function, it makes sense to normalize the strings in dict here.
      {:ok, declension} -> {:ok, %Adjective{declension: declension}}
      {:error, message} -> {:error, message}
    end
  end

  @doc """
  Like `regular` this function creates a regular Latin adjective. However,
  third declension adjectives created with this function will decline as
  consonant-stems instead of i-stems.

  Most third declension adjectives decline the same way as i-stem nouns.
  However, there are a handful of third declension adjectives with one
  termination that decline like consonant stem nouns. For example, the singular
  ablative of most third declension adjectives with one termination ends in -ī
  but consonant stem adjectives such as vetus, veteris end in -e.

  ## Example

      iex> {:ok, adjective} = Verba.Adjective.regular_c_stem(["vertus", "verteris"])
      {:ok,
        %Verba.Adjective{
          attr: %{c_stem: nil},
          declension: %{
            singular: %{
              genitive: %{
                feminine: ["verteris"],
                masculine: ["verteris"],
                neuter: ["verteris"]
              },
              nominative: %{
                feminine: ["vertus"],
                masculine: ["vertus"],
                neuter: ["vertus"]
              }
            }
          }
        }}
      iex> Verba.Adjective.decline(adjective, :plural, :genitive, :masculine)
      ["verterum"]

  """
  @spec regular_c_stem([String.t()]) :: {atom(), adjective()} | {atom(), String.t()}
  def regular_c_stem(dict = [_|_]) do
    case regular(dict) do
      {:ok, adjective} -> {:ok, %Adjective{adjective | attr: Map.put(adjective.attr, :c_stem, nil)}}
      {:error, message} -> {:error, message}
    end
  end

  @doc """
  Like `regular` this function creates a Latin adjective from a dictionary form
  However, first and second declension adjectives created with this function
  will decline with irregular singular dative and genitive forms.

  There are nine adjectives of the first and second declension group that
  decline differently from other first and second declension adjectives. These
  Tadjectives, commonly referred to by the acronym ŪNUS NAUTA are:

  ūllus, ūlla, ūllum
  nūllus, nūlla, nūllum
  uter, utra, utrum
  sōlus, sōla, sōlum
  neuter, neutra, neutrum
  alius, alia, aliud
  ūnus, ūna, ūnum
  tōtus, tōta, tōtum
  alter, altera, alterum

  Use this function when creating these adjectives.

  # Example
      iex> {:ok, adjective} = Verba.Adjective.irregular_ius_genitive(["ūllus", "ūlla", "ūllum"])
      {:ok,
        %Verba.Adjective{
          attr: %{ius_genitive_ending: nil},
          declension: %{
            singular: %{
              nominative: %{
                feminine: ["ūlla"],
                masculine: ["ūllus"],
                neuter: ["ūllum"]
              }
            }
          }
        }}
      iex> Verba.Adjective.decline(adjective, :singular, :genitive, :masculine)
      ["ūllīus"]
      iex> Verba.Adjective.decline(adjective, :singular, :dative, :masculine)
      ["ūllī"]

  """
  @spec irregular_ius_genitive([String.t()]) :: {atom(), adjective()} | {atom(), String.t()}
  def irregular_ius_genitive(dict = [_|_]) do
    case regular(dict) do
      {:ok, adjective} -> {:ok, %Adjective{adjective | attr: Map.put(adjective.attr, :ius_genitive_ending, nil)}}
      {:error, message} -> {:error, message}
    end
  end

  @doc """
  Returns the declension group and number of terminations of an `adjective()`
  or the dictionary form of an adjective. If the declension group cannot be
  determined, nil is returned.

  ## Examples

      iex> Verba.Adjective.group(["altus", "alta", "altum"])
      {:first_second, :three}
      iex> Verba.Adjective.group(["atrōx", "atrōcis"])
      {:third, :one}

      iex> {:ok, adjective} = Verba.Adjective.regular(["altus", "alta", "altum"])
      {:ok,
        %Verba.Adjective{
          attr: %{},
          declension: %{
            singular: %{
              nominative: %{
                feminine: ["alta"],
                masculine: ["altus"],
                neuter: ["altum"]
              }
            }
          }
        }}
      iex> Verba.Adjective.group(adjective)
      {:first_second, :three}

  """
  # Used when creating a regular adjective to determine if the provided
  # dictionary form matches a regular declension group. If it does, the group
  # and number of terminations are returned. If it doesn't, nil is returned.
  @spec group(adjective() | [String.t()]) :: {group(), termination()} | nil
  def group([masculine | [feminine | [neuter | []]]]) do
    cond do
      (String.ends_with?(masculine, "us") or String.ends_with?(masculine, "er")) and String.ends_with?(feminine, "a") and String.ends_with?(neuter, "um") -> {:first_second, :three}
      String.ends_with?(masculine, "er") and String.ends_with?(feminine, "is") and String.ends_with?(neuter, "e") -> {:third, :three}
      true -> nil
    end
  end

  def group([first | [second | []]]) do
    cond do
      String.ends_with?(first, "is") and String.ends_with?(second, "e") -> {:third, :two}
      String.ends_with?(second, "is") -> {:third, :one}
      true -> nil
    end
  end

  def group(adjective = %Adjective{}) do
    cond do
      first_and_second?(adjective) -> {:first_second, :three}
      third_one_ending?(adjective) -> {:third, :one}
      third_two_endings?(adjective) -> {:third, :two}
      third_three_endings?(adjective) -> {:third, :three}
      true -> nil
    end
  end

  def group(_) do
    nil
  end

  # Returns whether an adjective belongs to the first and second declension
  # group.
  @spec first_and_second?(adjective()) :: boolean()
  defp first_and_second?(adjective) do
    String.ends_with?(a_nom(adjective, :masculine), ["us", "er"]) and String.ends_with?(a_nom(adjective, :feminine), "a") and String.ends_with?(a_nom(adjective, :neuter), "um")
  end

  # Returns whether an adjective belongs to the third declension group and has
  # one termination. Third declension adjectives with one termination are the
  # only regular adjectives that store a genitive form so checking for that and
  # testing to ensure it has the correct ending is sufficient.
  @spec third_one_ending?(adjective()) :: boolean()
  defp third_one_ending?(%Adjective{declension: %{singular: %{genitive: %{masculine: [gen | _]}}}}) do
    String.ends_with?(gen, "is")
  end

  defp third_one_ending?(_) do
    false
  end

  # Returns whether an adjective belongs to the third declension group and has
  # two terminations.
  @spec third_two_endings?(adjective()) :: boolean()
  defp third_two_endings?(adjective) do
    String.ends_with?(a_nom(adjective, :masculine), "is") and String.ends_with?(a_nom(adjective, :feminine), "is") and String.ends_with?(a_nom(adjective, :neuter), "e")
  end

  # Returns whether an adjective belongs to the third declension group and has
  # three terminations.
  @spec third_three_endings?(adjective()) :: boolean()
  defp third_three_endings?(adjective) do
    String.ends_with?(a_nom(adjective, :masculine), "er") and String.ends_with?(a_nom(adjective, :feminine), "is") and String.ends_with?(a_nom(adjective, :neuter), "e")
  end

  @doc """
  Return the stem of an `adjective()` if it can be determined, otherwise nil is
  returned.

  ## Example

      iex> {:ok, adjective} = Verba.Adjective.regular(["altus", "alta", "altum"])
      {:ok,
        %Verba.Adjective{
          attr: %{},
          declension: %{
            singular: %{
              nominative: %{
                feminine: ["alta"],
                masculine: ["altus"],
                neuter: ["altum"]
              }
            }
          }
        }}
      iex> Verba.Adjective.stem(adjective)
      "alt"

  """
  @spec stem(adjective()) :: String.t() | nil
  def stem(adjective) do
    case group(adjective) do
      {group, termination} -> stem(adjective, group, termination)
      nil -> nil
    end
  end

  @spec stem(adjective(), group(), termination()) :: String.t()
  defp stem(adjective, :first_second, _) do
    a_nom(adjective, :feminine)
    |> String.trim_trailing("a")
  end

  defp stem(adjective, :third, :one) do
    a_gen(adjective, :masculine, :third, :one)
    |> String.trim_trailing("is")
  end

  # Creating the stem for third declension adjectives with two and three
  # terminations is the exact same.
  defp stem(adjective, :third, _) do
    a_nom(adjective, :feminine)
    |> String.trim_trailing("is")
  end

  # A convenience function to which to pass an adjective and list of endings
  # and get a list of the endings combined with the stem.
  @spec endings(adjective(), [String.t()]) :: [String.t()]
  defp endings(adjective, endings) do
    stem(adjective)
    |> Inflection.stem_with_endings(endings)
  end

  # The nominative form of an adjective varies depending on the declension
  # group, number of terminations, and the desired gender. This function takes
  # an adjective and a gender then returns the proper nominative form for that
  # gender.
  @spec all_nom(adjective(), Decline.grammatical_gender()) :: [String.t()]
  defp all_nom(%Adjective{declension: %{singular: %{nominative: %{masculine: masculine}}}}, :masculine) do
    masculine
  end

  defp all_nom(%Adjective{declension: %{singular: %{nominative: %{feminine: feminine}}}}, :feminine) do
    feminine
  end

  defp all_nom(%Adjective{declension: %{singular: %{nominative: %{neuter: neuter}}}}, :neuter) do
    neuter
  end

  @spec a_nom(adjective(), Decline.grammatical_gender()) :: String.t()
  defp a_nom(adjective, gender) do
    List.first(all_nom(adjective, gender))
  end

  # The genitive form of an adjective varies depending on the declension
  # group and, in the case of first and second declensoin adjectives, gender.
  # This function takes an adjective and a gender then returns the proper
  # genitive form.
  @spec a_gen(adjective(), Decline.grammatical_gender()) :: String.t()
  defp a_gen(adjective, gender) do
    {group, termination} = group(adjective)

    a_gen(adjective, gender, group, termination)
  end

  @spec a_gen(adjective(), Decline.grammatical_gender(), group(), termination()) :: String.t()
  defp a_gen(adjective, :feminine, :first_second, _) do
    stem(adjective, :first_second, :three) <> "ae"
  end

  defp a_gen(adjective, _, :first_second, _) do
    stem(adjective, :first_second, :three) <> "ī"
  end

  # Third declension adjectives with a single termination store their genitive
  # form. Furthermore, the singular genitive form is the same for all three
  # genders so this function fetches and returns the masculine form.
  defp a_gen(%Adjective{declension: %{singular: %{genitive: %{masculine: [genitive | _]}}}}, _, :third, :one) do
    genitive
  end

  defp a_gen(adjective, _, :third, termination) do
    stem(adjective, :third, termination) <> "is"
  end

  @spec all_gen(adjective(), Decline.grammatical_gender) :: [String.t()]
  defp all_gen(adjective, gender) do
    [a_gen(adjective, gender)]
  end

  @doc """
  Provides the declined form of an adjective for a specified number, case, and
  gender.

  Latin adjectives are declined for number, case, and gender. There are six
  cases; nominative, genitive, dative, accusative, ablative, and vocative; two
  two numbers; singular and plural; and three genders; masculine feminine, and
  neuter.

  ## Example

      iex> {:ok, adjective} = Verba.Adjective.regular(["altus", "alta", "altum"])
      {:ok,
        %Verba.Adjective{
          attr: %{},
          declension: %{
            singular: %{
              nominative: %{
                feminine: ["alta"],
                masculine: ["altus"],
                neuter: ["altum"]
              }
            }
          }
        }}
      iex> Verba.Adjective.decline(adjective, :plural, :genitive, :masculine)
      ["altōrum"]
  """
  @spec decline(adjective(), Decline.grammatical_number(), Decline.grammatical_case(), Decline.grammatical_gender()) :: [String.t()]
  def decline(adjective, grammatical_number, grammatical_case, grammatical_gender) do
    {group, termination} = group(adjective)

    decline(adjective, group, termination, grammatical_number, grammatical_case, grammatical_gender)
  end

  defp decline(adjective, _, _, :singular, :nominative, gender) do
    all_nom(adjective, gender)
  end

  # The singular genitive ending of a handful of first and second declension
  # adjectives is -īus for all three genders instead of the usual -ae for
  # feminine and -ī for masculine and neuter.
  defp decline(adjective = %Adjective{attr: %{ius_genitive_ending: _}}, :first_second, _, :singular, :genitive, _) do
    endings(adjective, ["īus"])
  end

  defp decline(adjective, _, _, :singular, :genitive, gender) do
    all_gen(adjective, gender)
  end

  #############################################################################
  # decline/6 for first and second declension adjectives.
  #############################################################################
  # The singular dative ending of a handful of first and second declension
  # adjectives is -ī for all three genders instead of the usual -ae for
  # feminine and -ō for masculine and neuter.
  defp decline(adjective = %Adjective{attr: %{ius_genitive_ending: _}}, :first_second, _, :singular, :dative, _) do
    endings(adjective, ["ī"])
  end

  defp decline(adjective, :first_second, _, :singular, :dative, :feminine) do
    endings(adjective, ["ae"])
  end

  defp decline(adjective, :first_second, _, :singular, :dative, _) do
    endings(adjective, ["ō"])
  end

  defp decline(adjective, :first_second, _, :singular, :accusative, :masculine) do
    endings(adjective, ["um"])
  end

  defp decline(adjective, :first_second, _, :singular, :accusative, :feminine) do
    endings(adjective, ["am"])
  end

  defp decline(adjective, :first_second, _, :singular, :accusative, gender) do
    all_nom(adjective, gender)
  end

  defp decline(adjective, :first_second, _, :singular, :ablative, :feminine) do
    endings(adjective, ["ā"])
  end

  defp decline(adjective, :first_second, _, :singular, :ablative, _) do
    endings(adjective, ["ō"])
  end

  # The singular vocative form of second declension masculine nouns is the same
  # as the nominative form if the singular mascline ends in -r. Otherwise it
  # is the stem plus an -e.
  defp decline(adjective, :first_second, _, :singular, :vocative, gender) when gender == :masculine do
    nom = all_nom(adjective, gender)

    cond do
      Decline.r_ending?(Enum.at(nom, 0)) -> nom
      true -> endings(adjective, ["e"])
    end
  end

  defp decline(adjective, :first_second, _, :singular, :vocative, gender) do
    all_nom(adjective, gender)
  end

  defp decline(adjective, :first_second, _, :plural, grammatical_case, :masculine) when grammatical_case in [:nominative, :vocative] do
    endings(adjective, ["ī"])
  end

  defp decline(adjective, :first_second, _, :plural, grammatical_case, :feminine) when grammatical_case in [:nominative, :vocative] do
    endings(adjective, ["ae"])
  end

  defp decline(adjective, :first_second, _, :plural, grammatical_case, :neuter) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(adjective, ["a"])
  end

  defp decline(adjective, :first_second, _, :plural, :genitive, :feminine) do
    endings(adjective, ["ārum"])
  end

  defp decline(adjective, :first_second, _, :plural, :genitive, _) do
    endings(adjective, ["ōrum"])
  end

  # The plural dative and ablative forms for first and second declension
  # adjectives are the same for all three genders.
  defp decline(adjective, :first_second, _, :plural, grammatical_case, _) when grammatical_case in [:dative, :ablative] do
    endings(adjective, ["īs"])
  end

  defp decline(adjective, :first_second, _, :plural, :accusative, :masculine) do
    endings(adjective, ["ōs"])
  end

  defp decline(adjective, :first_second, _, :plural, :accusative, :feminine) do
    endings(adjective, ["ās"])
  end

  #############################################################################
  # decline/6 for declensions common to all third declension adjectives.
  #############################################################################
  defp decline(adjective = %Adjective{attr: %{c_stem: _}}, :third, _, :singular, :dative, _) do
    endings(adjective, ["ī"])
  end

  defp decline(adjective = %Adjective{attr: %{c_stem: _}}, :third, _, :singular, :ablative, _) do
    endings(adjective, ["e"])
  end

  defp decline(adjective, :third, _, :singular, grammatical_case, _) when grammatical_case in [:dative, :ablative] do
    endings(adjective, ["ī"])
  end

  defp decline(adjective, :third, _, :singular, :accusative, :neuter) do
    all_nom(adjective, :neuter)
  end

  defp decline(adjective, :third, _, :singular, :accusative, _) do
    endings(adjective, ["em"])
  end

  defp decline(adjective, :third, _, :singular, :vocative, gender) do
    all_nom(adjective, gender)
  end

  defp decline(adjective, :third, _, :plural, grammatical_case, :neuter) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(adjective, ["ia"])
  end

  defp decline(adjective, :third, _, :plural, grammatical_case, _) when grammatical_case in [:nominative, :vocative] do
    endings(adjective, ["ēs"])
  end

  defp decline(adjective = %Adjective{attr: %{c_stem: _}}, :third, _, :plural, :genitive, _) do
    endings(adjective, ["um"])
  end


  defp decline(adjective, :third, _, :plural, :genitive, _) do
    endings(adjective, ["ium"])
  end

  defp decline(adjective, :third, _, :plural, grammatical_case, _) when grammatical_case in [:dative, :ablative] do
    endings(adjective, ["ibus"])
  end

  defp decline(adjective = %Adjective{attr: %{c_stem: _}}, :third, _, :plural, :accusative, _) do
    endings(adjective, ["ēs"])
  end

  defp decline(adjective, :third, _, :plural, :accusative, _) do
    endings(adjective, ["ēs", "īs"])
  end

end
