defmodule Verba.Decline do
  @type grammatical_number() :: :singular | :plural
  @type grammatical_gender() :: :masculine | :feminine | :neuter
  @type grammatical_case() :: :nominative | :genitive | :dative | :accusative | :ablative | :vocative | :locative

  # Checks if a regular second declension noun has an -ius ending in the case
  # of masculine and feminine nouns or an -ium ending in the case of neuter
  # nouns.
  @spec ius_ending?(String.t()) :: boolean()
  def ius_ending?(nominative) do
    cond do
      String.ends_with?(nominative, ["ius", "ium"]) -> true
      true -> false
    end
  end

  # Checks if a regular second declension masculine noun has an -r ending. Such
  # nouns decline in a slightly different way from other second declension
  # masculine nouns.
  @spec r_ending?(String.t()) :: boolean()
  def r_ending?(nominative) do
    cond do
      String.ends_with?(nominative, "r") -> true
      true -> false
    end
  end

  defp syllables([], acc) do
    acc
  end

  defp syllables([final | []], acc) do
    case vowel?(final) do
      true -> acc + 1
      false -> acc
    end
  end

  defp syllables([first | tail], acc) do
    case vowel?(first) do
      true ->
        [second | rest] = tail

        case diphthong?(first, second) do
          true -> syllables(rest, acc + 1) # If the first and next grapheme forms a diphthong, skip checking the second letter.
          false -> syllables(tail, acc + 1) # Otherwise check the second letter.
        end
      false -> syllables(tail, acc)
    end
  end

  @doc """
  Returns the number of syllables in a Latin word.

  ## Example

      iex> import Verba.Decline
      iex> syllables("animalia")
      5
      iex> syllables("saepe")
      2
  """
  @spec syllables(String.t()) :: integer()
  def syllables(word) do
    syllables(String.graphemes(word), 0)
  end

  @doc """
  Returns whether or not a grapheme is a vowel.

  ##Example

      iex> import Verba.Decline
      iex> vowel?("a")
      true
      iex> vowel?("ō")
      true
      iex> vowel?("s")
      false
  """
  @spec vowel?(String.grapheme()) :: boolean()
  def vowel?(grapheme) when grapheme in ["a", "ā", "e", "ē", "i", "ī", "o", "ō", "u", "ū"] do
    true
  end

  def vowel?(_) do
    false
  end

  @doc """
  Returns whether or not two graphemes form a diphthong in Classical Latin.

  ## Example

      iex> import Verba.Decline
      iex> diphthong?("a", "e")
      true
      iex> diphthong?("a", "o")
      false
  """
  @spec diphthong?(String.grapheme(), String.grapheme()) :: boolean()
  def diphthong?(first, second) when first in ["a", "ā"] do
    second in ["e", "ē", "u", "ū"]
  end

  def diphthong?(first, second) when first in ["e", "ē"] do
    second in ["i", "ī", "u", "ū"]
  end

  def diphthong?(first, second) when first in ["o", "ō"] do
    second in ["e", "ē"]
  end

  def diphthong?(first, second) when first in ["u", "ū"] do
    second in ["i", "ī"]
  end

  def diphthong?(_, _) do
    false
  end

  @doc """
  This function attempts to determine whether a passed in third declension noun
  or adjective is a pure i-stem. The rules checked against are from Wheelock's
  Latin and are as follows:

  1. Masculine and feminine nouns and adjectives that end in -is or -es that
     have the same number of syllables in both their nominative and genitive
     forms.

  2. Neuter nouns and adjectives that end in -al, -ar, or -e.

  ## Warning

  This function likely won't be 100 percent accurate. Human language is always
  vague. If this function ends up providing more false positives than I'm
  comfortable with, I will likely decricate it and instead store whether or not
  a noun or adjective is an i-stem in the word database. Don't depend on this
  function too much as it may go away.


  """
  @spec i_stem_parisyllabic?(String.t(), String.t(), grammatical_gender()) :: boolean()
  def i_stem_parisyllabic?(nom, gen, gender) when gender in [:masculine, :feminine] do
    # The parisyllabic rule states that masculing and feminine nouns and
    # adjectives whose nominative case ends with -is or -es and whose
    # nominative and genitive cases have the same number of syllables are
    # i-stem.
    case String.ends_with?(nom, ["is", "ēs"]) do
      true -> syllables(nom) == syllables(gen)
      false -> false
    end
  end

  def i_stem_parisyllabic?(nom, _, :neuter) do
    String.ends_with?(nom, ["al", "ar", "e"])
  end

  @doc """
  This function attempts to determine whether a passed in third declension noun
  or adjective is a mixed i-stem. The rules checked against are from Wheelock's
  Latin and are as follows:

  1. Masculine and feminine nouns and adjectives that end in -s or -x with a
     stem ending in two consonants.

  ## Warning

  This function likely won't be 100 percent accurate. Human language is always
  vague. If this function ends up providing more false positives than I'm
  comfortable with, I will likely depricate it and instead store whether or not
  a noun or adjective is an i-stem in the word database. Don't depend on this
  function too much as it may go away.
  """
  @spec i_stem_two_consonant_ending?(String.t(), String.t(), grammatical_gender()) :: boolean()
  def i_stem_two_consonant_ending?(nom, stem, gender) when gender in [:masculine, :feminine] do
    cond do
      String.ends_with?(nom, ["s", "x"]) and syllables(nom) == 1 and String.slice(stem, -2, 2) |> String.graphemes() |> Enum.all?(fn c -> vowel?(c) == false end) -> true
      String.ends_with?(nom, ["ns", "rs"]) and syllables(nom) > 1 -> true
      true -> false
    end
  end

  def i_stem_two_consonant_ending?(_, _, :neuter) do
    false
  end

  # Determines if a noun is an i-stem.
  @spec i_stem?(String.t(), String.t(), String.t(), grammatical_gender()) :: boolean()
  def i_stem?(nom, gen, stem, gender) do
    cond do
      i_stem_parisyllabic?(nom, gen, gender) -> true
      i_stem_two_consonant_ending?(nom, stem, gender) -> true
      true -> false
    end
  end

  @doc """
  Normalizes a declension into Unicode composed form.

  Declensions can either refer to a list of strings (for example, the singular
  nominative and genitive forms of a noun passed into `Verba.Noun.regular`) or
  a Map containing either lists of strings or nested Maps that ultimately
  contain lists of strings (for example, the Map of irregular declined forms
  passed to `Verba.Noun.irregular`).

  ## Examples

  When called with a list of strings, this function returns a list of Unicode
  normalized composed form strings.

      iex> bellum = Verba.Decline.normalize_declension(["bellum", "bellī"])
      ["bellum", "bellī"]
      iex> Enum.at(bellum, 0) == :unicode.characters_to_nfc_binary("bellum")
      true
      iex> Enum.at(bellum, 1) == :unicode.characters_to_nfc_binary("bellī")
      true

  When called with a Map ultimately containing lists of strings, this function
  returns an equivalent map with all of the strings in Unicode normalization
  composed form.

      iex> porta = %{singular: %{nominative: ["porta"], genitive: ["portae"]}, plural: %{accusative: ["portās"], genitive: ["portārum"]}}
      %{
        plural: %{accusative: ["portās"], genitive: ["portārum"]},
        singular: %{genitive: ["portae"], nominative: ["porta"]}
      }
      iex> normalized = Verba.Decline.normalize_declension(porta)
      %{
        plural: %{accusative: ["portās"], genitive: ["portārum"]},
        singular: %{genitive: ["portae"], nominative: ["porta"]}
      }
      iex> Enum.at(Map.get(normalized.singular, :nominative), 0) == :unicode.characters_to_nfc_binary("porta")
      true
      iex> Enum.at(Map.get(normalized.singular, :genitive), 0) == :unicode.characters_to_nfc_binary("portae")
      true
      iex> Enum.at(Map.get(normalized.plural, :accusative), 0) == :unicode.characters_to_nfc_binary("portās")
      true
      iex> Enum.at(Map.get(normalized.plural, :genitive), 0) == :unicode.characters_to_nfc_binary("portārum")
      true

  """
  @spec normalize_declension([String.t()] | map()) :: [String.t()] | map()
  # When called with a Map, this function will get the value stored by the
  # first key of the map and call itself with that value. If the value is
  # another Map, the same process is repeated. If the value is a list of
  # strings, a new list of normalized strings is returned.
  #
  # This function will therefore unwind nested Maps, normalize any contained
  # strings, and insert those normalized strings into a new Map with the exact
  # same key/value pairs.
  def normalize_declension(declension = %{}) when map_size(declension) >= 1 do
    [key | _] = Map.keys(declension) # Get the first key from the declension Map.

    %{key => normalize_declension(Map.get(declension, key))} # Get the value stored in the first key of the declension map, call this function with that value, and add that value to the new empty Map.
    |> Map.merge(normalize_declension(Map.delete(declension, key))) # Drop the value stored in the first key of the declension map, call this function with the remaining Map, and merge the returned Map with the Map created in this function.
  end

  def normalize_declension(declension = %{}) when map_size(declension) == 0 do
    %{}
  end

  def normalize_declension([]) do
    []
  end

  def normalize_declension([head | tail]) do
    [:unicode.characters_to_nfc_binary(head) | normalize_declension(tail)]
  end

end
