defmodule Verba.Noun do
  alias Verba.Noun, as: Noun

  import Verba.Decline
  alias Verba.Inflection, as: Inflection

  @typedoc """
  Regular Latin nouns belong to one of five declension groups.
  """
  @type group() :: :first | :second | :third | :fourth | :fifth

  # declension: stores all of the needed declined forms. For regular nouns,
  # just the singular nominative and genitive forms are stored. For irregular
  # nouns every declined form is stored. For indeclinable nouns just the
  # singular nominative is stored.
  #
  # gender: stores the grammatical gender of the noun.
  #
  # attr: stores a list of special attributes if applicable. For example, third
  # declension nouns decline differently in a handful of cases if they fall
  # into the i-stem category. If a third declension adjective is an i-stem,
  # attr will contain a the key i_stem: with a value of nil.
  defstruct declension: %{}, gender: :masculine, attr: %{}

  @opaque noun() :: %Noun{}

  @doc """
  Creates a regular Latin noun.

  The dictionary form of a regular Latin noun consistes of its singular
  nominative and genitive forms. For example, for the second declension neuter
  noun bellum (war) the dictionary form would be bellum, bellī. Latin nouns are
  also gendered. Latin, like German, has three genders: masculine, feminine,
  and neuter.

  This function takes a list of singular nominative and genitive forms (fourth
  declension neuter nouns have two genitive forms) and a gender. With this
  information, the correct declined form for any case can be determined.

  ## Warning

  Third declension nouns can be either consonant-stem or i-stem. Third
  declension nouns created with this function will decline as consonant-stem.
  When creating a third declension i-stem noun, use `regular_i_stem` instead.

  ## Example

      iex> nom = ["bellum"]
      ["bellum"]
      iex> gen = ["bellī"]
      ["bellī"]
      iex> Verba.Noun.regular(nom, gen, :neuter)
      {:ok,
        %Verba.Noun{
          attr: %{},
          declension: %{singular: %{genitive: ["bellī"], nominative: ["bellum"]}},
          gender: :neuter
      }}
  """
  @spec regular([String.t()], [String.t()], Decline.grammatical_gender()) :: {atom(), noun()} | {atom(), String.t()}
  def regular(nom = [_|_], gen = [_|_], gender) do
    case group(Enum.at(gen, 0)) do
      nil -> {:error, "The ending of the singular genitive form does not match that of any declension group."}
      _ -> {:ok, %Noun{declension: %{:singular => %{:nominative => normalize_declension(nom), :genitive => normalize_declension(gen)}}, gender: gender}}
    end
  end

  @doc """
  Like `regular` this function creates a regular Latin noun. However, third
  declension nouns created with this function will decline as i-stems instead
  of consonant stems.

  ## Examples

      iex> nom = ["animal"]
      ["animal"]
      iex> gen = ["animālis"]
      ["animālis"]
      iex> {_, noun} = Verba.Noun.regular_i_stem(nom, gen, :neuter)
      {:ok,
        %Verba.Noun{
          attr: %{i_stem: nil},
          declension: %{singular: %{genitive: ["animālis"], nominative: ["animal"]}},
          gender: :neuter
      }}
      iex> Verba.Noun.decline(noun, :plural, :genitive)
      ["animālium"]
  """
  @spec regular_i_stem([String.t()], [String.t()], Decline.grammatical_gender()) :: {atom(), noun()} | {atom(), String.t()}
  def regular_i_stem(nom = [_|_], gen = [_|_], gender) do
    case regular(nom, gen, gender) do
      {:ok, noun} -> {:ok, %Noun{noun | attr: Map.put(noun.attr,  :i_stem , nil)}}
      {:error, message} -> {:error, message}
    end
  end

  @doc """
  Creates an irregular Latin noun.

  Latin nouns can contain declensions that do not follow the rules of any
  regular declension group. These nouns are, not surprisingly, referred to as
  irregular.

  This function provides a method for creating irregular nouns. Unlike
  [`regular`], which takes the singular nominative and genitive forms, this
  function takes a Map containing all of the irregular declined forms, the
  stem, and the declension group. If a declined form isn't present in the Map,
  it is assumed to be regular and will be declined according to the regular
  Latin rules for the passed in declension group.

  For example, if the Map contains values for the singular dative form but not
  the singular ablative form, the singular dative form will be declined as it
  appears in the Map but the singular ablative form will be declined the same
  as a noun created using `regular`.

  ## Warning

  Third declension nouns can be either consonant-stem or i-stem. Third
  declension nouns created with this function will decline as consonant-stem.
  When creating a third declension i-stem noun, use `irregular_i_stem` instead.

  ## Example

  vīs, vīs is an irregular noun with an irregular singular genitive form. The
  stem is vīr- but none of the singular cases use. Moreover, the plural cases
  decline like a third declension i-stem. So all six singular forms must be
  provided as do the plural genitive and accusative forms. The rest of the
  forms can be determined by combining the provided stem with the regular third
  declension endings.

  This example shows how to create the Map containing the irregular declined
  forms and demonstrates two declensions. The first declension is regular, the
  second is irregular.

      iex> irregular = %{singular: %{nominative: ["vīs"], genitive: ["vīs"], dative: ["vī"], accusative: ["vim"], ablative: ["vī"], vocative: ["vīs"]}, plural: %{genitive: ["vīrium"], accusative: ["vīrēs", "vīrīs"]}}
      %{
        plural: %{accusative: ["vīrēs", "vīrīs"], genitive: ["vīrium"]},
        singular: %{
        ablative: ["vī"],
        accusative: ["vim"],
        dative: ["vī"],
        genitive: ["vīs"],
        nominative: ["vīs"],
        vocative: ["vīs"]
        }
      }
      iex> {:ok, noun} = Verba.Noun.irregular(irregular, :feminine, "vīr", :third)
      {:ok,
        %Verba.Noun{
          attr: %{group: :third, irregular: nil, stem: "vīr"},
          declension: %{
            plural: %{accusative: ["vīrēs", "vīrīs"], genitive: ["vīrium"]},
            singular: %{
              ablative: ["vī"],
              accusative: ["vim"],
              dative: ["vī"],
              genitive: ["vīs"],
              nominative: ["vīs"],
              vocative: ["vīs"]
            }
          },
          gender: :feminine
      }}
      iex> Verba.Noun.decline(noun, :singular, :genitive)
      ["vīs"]
      iex> Verba.Noun.decline(noun, :plural, :genitive)
      ["vīrium"]
      iex> Verba.Noun.decline(noun, :plural, :dative)
      ["vīribus"]

  """
  @spec irregular(%{}, Decline.grammatical_gender(), String.t(), group()) :: {atom(), noun()} | {atom(), String.t()}
  def irregular(declension, gender, stem, group) do
    # The declension Map must contain singular nominative and genitive forms at
    # the very least.
    case declension do
      %{:singular =>
        %{:nominative => [_|_], :genitive => [_|_]}
      } -> {:ok, %Noun{declension: normalize_declension(declension), gender: gender, attr: %{irregular: nil, stem: :unicode.characters_to_nfc_binary(stem), group: group}}}
      _ -> {:error, "The declension Map doesn't contain singular nominative and genitive forms."}
    end
  end

  @doc """
  Like `irregular` this function creates an irregular Latin noun. However,
  third declension nouns created with this function will decline as i-stems
  instead of consonant stems.

  ## Example

  vīs, vīs, which is also used in the example for [`irregular`] is actually an
  i-stem. In the example for `irregular` the i-stem cases, namely the plural
  genitive and plural accusative, were added to the declension Map. In this
  example they are omitted since irregular nouns created with this function
  decline as i-stems.

      iex> irregular = %{singular: %{nominative: ["vīs"], genitive: ["vīs"], dative: ["vī"], accusative: ["vim"], ablative: ["vī"], vocative: ["vīs"]}}
      %{
        singular: %{
        ablative: ["vī"],
        accusative: ["vim"],
        dative: ["vī"],
        genitive: ["vīs"],
        nominative: ["vīs"],
        vocative: ["vīs"]
        }
      }
      iex> {:ok, noun} = Verba.Noun.irregular_i_stem(irregular, :feminine, "vīr", :third)
      {:ok,
        %Verba.Noun{
          attr: %{group: :third, irregular: nil, stem: "vīr", i_stem: nil},
          declension: %{
            singular: %{
              ablative: ["vī"],
              accusative: ["vim"],
              dative: ["vī"],
              genitive: ["vīs"],
              nominative: ["vīs"],
              vocative: ["vīs"]
            }
          },
          gender: :feminine
      }}
      iex> Verba.Noun.decline(noun, :singular, :genitive)
      ["vīs"]
      iex> Verba.Noun.decline(noun, :plural, :genitive)
      ["vīrium"]
      iex> Verba.Noun.decline(noun, :plural, :dative)
      ["vīribus"]

  """
  @spec irregular_i_stem(%{}, Decline.grammatical_gender(), String.t(), group()) :: {atom(), noun()} | {atom(), String.t()}
  def irregular_i_stem(declension, gender, stem, group) do
    case irregular(declension, gender, stem, group) do
      {:ok, noun} -> {:ok, %Noun{noun | attr: Map.put(noun.attr, :i_stem, nil)}}
      {:error, message} -> {:error, message}
    end
  end

  @doc """
  Unlike other Latin nouns, indeclinable nouns only have one form and don't
  belong to a declension group.

  ## Example

      iex> {:ok, noun} = Verba.Noun.indeclinable(["nīl"], :neuter)
      {:ok,
        %Verba.Noun{
          attr: %{indeclinable: nil},
          declension: %{singular: %{nominative: ["nīl"]}},
          gender: :neuter
      }}
      iex> Verba.Noun.decline(noun, :singular, :nominative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :singular, :genitive)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :singular, :dative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :singular, :accusative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :singular, :ablative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :singular, :vocative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :plural, :nominative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :plural, :genitive)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :plural, :dative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :plural, :accusative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :plural, :ablative)
      ["nīl"]
      iex> Verba.Noun.decline(noun, :plural, :vocative)
      ["nīl"]

  """
  @spec indeclinable([String.t()], Decline.grammatical_gender()) :: {atom(), noun()} | {atom(), String.t()}
  def indeclinable(nom, gender) when length(nom) >= 1 do
    {:ok, %Noun{declension: %{:singular => %{:nominative => normalize_declension(nom)}}, gender: gender, attr: %{indeclinable: nil}}}
  end

  @doc """
  Takes a `noun` or the singular genitive form of a Latin noun and returns the
  declension group. In the cases of indeclinable nouns, which don't belong to a
  declension group, and singular genitive forms whose declension group cannot
  be determined, `nil` is returned.

  ## Example

      iex> {:ok, noun} = Verba.Noun.regular(["bellum"], ["bellī"], :neuter)
      {:ok,
        %Verba.Noun{
          attr: %{},
          declension: %{singular: %{genitive: ["bellī"], nominative: ["bellum"]}},
          gender: :neuter
      }}
      iex> Verba.Noun.group(noun)
      :second

      iex> Verba.Noun.group("bellī")
      :second
  """
  @spec group(String.t() | noun()) :: group() | nil
  def group(%Noun{attr: %{indeclinable: _}}) do
    nil
  end

  def group(%Noun{attr: %{irregular: _, group: group}}) do
    group
  end

  def group(noun = %Noun{}) do
    group(a_gen(noun))
  end

  def group(gen) do
    cond do
      String.ends_with?(gen, ["ēī", "eī"]) -> :fifth # If this clause doesn't appear before the clause for second declension nouns, fifth declension nouns will be identified as second.
      String.ends_with?(gen, ["ūs", "ū"]) -> :fourth # Singular genitive forms of neuter fourth declension nouns can end in -ūs or -ū.
      String.ends_with?(gen, "is") -> :third
      String.ends_with?(gen, "ī") -> :second
      String.ends_with?(gen, "ae") -> :first
      true -> nil
    end
  end

  @doc """
  Takes a `noun` or the singular nominative and genitive forms of a Latin noun
  and returns the noun's stem. If this function is called with an indeclinable
  `noun` or a pair of singular nominative and genitive forms whose stem cannot
  be determined, `nil` is returned.

  ## Examples

      iex> {:ok, noun} = Verba.Noun.regular(["bellum"], ["bellī"], :neuter)
      {:ok,
        %Verba.Noun{
          attr: %{},
          declension: %{singular: %{genitive: ["bellī"], nominative: ["bellum"]}},
          gender: :neuter
      }}
      iex> Verba.Noun.stem(noun)
      "bell"

      iex> Verba.Noun.stem("bellum", "bellī")
      "bell"

  """
  @spec stem(noun()) :: String.t() | atom()
  def stem(%Noun{attr: %{indeclinable: _}}) do
    nil
  end

  def stem(%Noun{attr: %{irregular: _, stem: stem}}) do
    stem
  end

  def stem(noun = %Noun{}) do
    stem(a_nom(noun), a_gen(noun), group(noun))
  end

  @spec stem(String.t(), String.t()) :: String.t() | atom()
  def stem(nom, gen) do
    stem(nom, gen, group(gen))
  end

  @spec stem(String.t(), String.t(), group()) :: String.t() | atom()
  defp stem(_, gen, :first) do
    String.trim_trailing(gen, "ae")
  end

  defp stem(nom, gen, :second) do
    # Masculine and feminine second declension nouns ending in -ius and neuter
    # ones ending in -ium have an -iī singular genitive ending that must be
    # removed to get the stem. All others have an -ī ending.
    cond do
      ius_ending?(nom) -> String.trim_trailing(gen, "iī")
      String.ends_with?(gen, "ī") -> String.trim_trailing(gen, "ī")
    end
  end

  defp stem(_, gen, :third) do
    String.trim_trailing(gen, "is")
  end

  defp stem(_, gen, :fourth) do
    cond do
      String.ends_with?(gen, "ūs") -> String.trim_trailing(gen, "ūs")
      String.ends_with?(gen, "ū") -> String.trim_trailing(gen, "ū")
    end
  end

  defp stem(_, gen, :fifth) do
    cond do
      String.ends_with?(gen, "ēī") -> String.trim_trailing(gen, "ēī")
      String.ends_with?(gen, "eī") -> String.trim_trailing(gen, "eī")
    end
  end

  # stem/1 and stem/2 call this function with group/1 as the third argument.
  # If stem/2 is called, group/1 may return nil. If that happens, then no stem
  # should be returned.
  defp stem(_, _, nil) do
    nil
  end

  # A convenience function to which to pass a noun and list of endings and get
  # a list of the endings combined with the noun's stem.
  defp endings(noun, endings) do
    # Since noun() types are created with functions that ensure a stem can be
    # can be obtained, there is no need to check for a nil value here.
    stem(noun)
    |> Inflection.stem_with_endings(endings)
  end

  # When I need a nominative form but it doesn't matter which one.
  @spec a_nom(noun()) :: String.t()
  defp a_nom(noun) do
    all_nom(noun)
    |> Enum.at(0)
  end

  # When I need a genitive form but it doesn't matter which one.
  @spec a_gen(noun()) :: String.t()
  defp a_gen(noun) do
    all_gen(noun)
    |> Enum.at(0)
  end

  # When I need all nominative forms.
  @spec all_nom(noun()) :: [String.t()]
  defp all_nom(noun) do
    Map.get(noun.declension, :singular)
    |> Map.get(:nominative)
  end

  # When I need all genitive forms.
  @spec all_gen(noun()) :: [String.t()]
  defp all_gen(noun) do
    Map.get(noun.declension, :singular)
    |> Map.get(:genitive)
  end

  @doc """
  Declines a regular Latin noun for case and number.

  ## Example

      iex> {:ok, noun} = Verba.Noun.regular(["bellum"], ["bellī"], :neuter)
      {:ok,
        %Verba.Noun{
          attr: %{},
          declension: %{singular: %{genitive: ["bellī"], nominative: ["bellum"]}},
          gender: :neuter
      }}
      iex> Verba.Noun.decline(noun, :plural, :genitive)
      ["bellōrum"]
  """
  @spec decline(noun(), Decline.grammatical_number(), Decline.grammatical_case()) :: [String.t()]
  def decline(noun = %Noun{attr: %{indeclinable: _}}, _, _) do
    all_nom(noun)
  end

  # Since irregular nouns store any irregular forms in their declension Map,
  # decline/3 must first check to see if the desired declined form is stored.
  # If it is, it will be returned. If it isn't, the declined from will be
  # created the same way regular noun declensions are created.
  def decline(%Noun{declension: %{singular: %{nominative: declension}}, attr: %{irregular: _}}, :singular, :nominative) do
    declension
  end

  def decline(%Noun{declension: %{singular: %{genitive: declension}}, attr: %{irregular: _}}, :singular, :genitive) do
    declension
  end

  def decline(%Noun{declension: %{singular: %{dative: declension}}, attr: %{irregular: _}}, :singular, :dative) do
    declension
  end

  def decline(%Noun{declension: %{singular: %{accusative: declension}}, attr: %{irregular: _}}, :singular, :accusative) do
    declension
  end

  def decline(%Noun{declension: %{singular: %{ablative: declension}}, attr: %{irregular: _}}, :singular, :ablative) do
    declension
  end

  def decline(%Noun{declension: %{singular: %{vocative: declension}}, attr: %{irregular: _}}, :singular, :vocative) do
    declension
  end

  def decline(%Noun{declension: %{plural: %{nominative: declension}}, attr: %{irregular: _}}, :plural, :nominative) do
    declension
  end

  def decline(%Noun{declension: %{plural: %{genitive: declension}}, attr: %{irregular: _}}, :plural, :genitive) do
    declension
  end

  def decline(%Noun{declension: %{plural: %{dative: declension}}, attr: %{irregular: _}}, :plural, :dative) do
    declension
  end

  def decline(%Noun{declension: %{plural: %{accusative: declension}}, attr: %{irregular: _}}, :plural, :accusative) do
    declension
  end

  def decline(%Noun{declension: %{plural: %{ablative: declension}}, attr: %{irregular: _}}, :plural, :ablative) do
    declension
  end

  def decline(%Noun{declension: %{plural: %{vocative: declension}}, attr: %{irregular: _}}, :plural, :vocative) do
    declension
  end

  def decline(noun, grammatical_num, grammatical_case) do
    decline(noun, group(noun), grammatical_num, grammatical_case)
  end

  # Since every regular noun stores the singular nominative and genitive forms,
  # decline/3 returns them directly regardless of a noun's declension group.
  @spec decline(noun(), group(), Decline.grammatical_number, Decline.grammatical_case) :: [String.t()]
  defp decline(noun, _, :singular, :nominative) do
    all_nom(noun)
  end

  defp decline(noun, _, :singular, :genitive) do
    all_gen(noun)
  end

  #############################################################################
  # decline/4 for first declension nouns.
  #############################################################################

  defp decline(noun, :first, :singular, :vocative) do
    all_nom(noun)
  end

  defp decline(noun, :first, :singular, :dative) do
    endings(noun, ["ae"])
  end

  defp decline(noun, :first, :singular, :accusative) do
    endings(noun, ["am"])
  end

  defp decline(noun, :first, :singular, :ablative) do
    endings(noun, ["ā"])
  end

  defp decline(noun, :first, :plural, grammatical_case) when grammatical_case in [:nominative, :vocative] do
    endings(noun, ["ae"])
  end

  defp decline(noun, :first, :plural, :genitive) do
    endings(noun, ["ārum"])
  end

  defp decline(noun, :first, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    endings(noun, ["īs"])
  end

  defp decline(noun, :first, :plural, :accusative) do
    endings(noun, ["ās"])
  end

  #############################################################################
  # decline/4 for second declension nouns.
  #############################################################################

  # The accusative and vocative singular of neuter second declension nouns are
  # the same as thenominative.
  defp decline(noun = %Noun{gender: :neuter}, :second, :singular, grammatical_case) when grammatical_case in [:accusative, :vocative] do
    all_nom(noun)
  end

  defp decline(noun = %Noun{gender: :neuter}, :second, :singular, grammatical_case) when grammatical_case in [:dative, :ablative] do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iō"])
      true -> endings(noun, ["ō"])
    end
  end

  defp decline(noun = %Noun{gender: :neuter}, :second, :plural, grammatical_case) when grammatical_case in [:nominative, :accusative, :vocative] do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["ia"])
      true -> endings(noun, ["a"])
    end
  end

  defp decline(noun = %Noun{gender: :neuter}, :second, :plural, :genitive) do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iōrum"])
      true -> endings(noun, ["ōrum"])
    end
  end

  defp decline(noun = %Noun{gender: :neuter}, :second, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iīs"])
      true -> endings(noun, ["īs"])
    end
  end

  # Decline nouns that aren't neuter.
  defp decline(noun, :second, :singular, grammatical_case) when grammatical_case in [:dative, :ablative] do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iō"])
      true -> endings(noun, ["ō"])
    end
  end

  defp decline(noun, :second, :singular, :accusative)  do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["ium"])
      true -> endings(noun, ["um"])
    end
  end

  defp decline(noun, :second, :singular, :vocative)  do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["ī"])
      r_ending?(a_nom(noun)) -> all_nom(noun)
      true -> endings(noun, ["e"])
    end
  end

  defp decline(noun, :second, :plural, grammatical_case) when grammatical_case in [:nominative, :vocative] do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iī"])
      true -> endings(noun, ["ī"])
    end
  end

  defp decline(noun, :second, :plural, :genitive) do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iōrum"])
      true -> endings(noun, ["ōrum"])
    end
  end

  defp decline(noun, :second, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iīs"])
      true -> endings(noun, ["īs"])
    end
  end

  defp decline(noun, :second, :plural, :accusative) do
    cond do
      ius_ending?(a_nom(noun)) -> endings(noun, ["iōs"])
      true -> endings(noun, ["ōs"])
    end
  end

  #############################################################################
  # decline/4 for third declension nouns.
  #############################################################################
  defp decline(noun = %Noun{gender: :neuter}, :third, :singular, :dative) do
    endings(noun, ["ī"])
  end

  # The accusative and vocative singular of neuter third declension nouns are
  # the same as the nominative.
  defp decline(noun = %Noun{gender: :neuter}, :third, :singular, grammatical_case) when grammatical_case in [:accusative, :vocative] do
    all_nom(noun)
  end

  defp decline(noun = %Noun{gender: :neuter, attr: %{i_stem: _}}, :third, :singular, :ablative) do
    endings(noun, ["ī"])
  end

  defp decline(noun = %Noun{gender: :neuter}, :third, :singular, :ablative) do
    endings(noun, ["e"])
  end

  defp decline(noun = %Noun{gender: :neuter, attr: %{i_stem: _}}, :third, :plural, grammatical_case) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(noun, ["ia"])
  end

  defp decline(noun = %Noun{gender: :neuter}, :third, :plural, grammatical_case) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(noun, ["a"])
  end

  defp decline(noun = %Noun{attr: %{i_stem: _}}, :third, :plural, :genitive) do
    endings(noun, ["ium"])
  end

  defp decline(noun, :third, :plural, :genitive) do
    endings(noun, ["um"])
  end

  defp decline(noun = %Noun{gender: :neuter}, :third, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    endings(noun, ["ibus"])
  end

  # Declining masculine and feminine third declension nouns.
  defp decline(noun, :third, :singular, :dative) do
    endings(noun, ["ī"])
  end

  defp decline(noun = %Noun{attr: %{i_stem: _}}, :third, :singular, :accusative) do
    endings(noun, ["em", "im"])
  end

  defp decline(noun, :third, :singular, :accusative) do
    endings(noun, ["em"])
  end

  defp decline(noun, :third, :singular, :ablative) do
      endings(noun, ["e"])
  end

  # The vocative singular of masculine and feminine third declension nouns are
  # the same as the nominative.
  defp decline(noun, :third, :singular, :vocative)  do
    all_nom(noun)
  end

  defp decline(noun, :third, :plural, grammatical_case) when grammatical_case in [:nominative, :vocative] do
    endings(noun, ["ēs"])
  end

  defp decline(noun = %Noun{attr: %{i_stem: _}}, :third, :plural, :accusative) do
    endings(noun, ["ēs", "īs"])
  end

  defp decline(noun, :third, :plural, :accusative) do
    endings(noun, ["ēs"])
  end

  defp decline(noun, :third, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    endings(noun, ["ibus"])
  end

  #############################################################################
  # decline/4 for fourth declension nouns.
  #############################################################################

  defp decline(noun = %Noun{gender: :neuter}, :fourth, :singular, grammatical_case) when grammatical_case in [:accusative, :ablative, :vocative] do
    all_nom(noun)
  end

  defp decline(noun = %Noun{gender: :neuter}, :fourth, :singular, :dative) do
    endings(noun, ["ū", "uī"])
  end

  defp decline(noun = %Noun{gender: :neuter}, :fourth, :plural, grammatical_case) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(noun, ["ua"])
  end

  defp decline(noun = %Noun{gender: :neuter}, :fourth, :plural, :genitive) do
    endings(noun, ["uum"])
  end

  defp decline(noun = %Noun{gender: :neuter}, :fourth, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    endings(noun, ["ibus"])
  end

  defp decline(noun, :fourth, :singular, :vocative) do
    all_nom(noun)
  end

  defp decline(noun, :fourth, :singular, :dative) do
    endings(noun, ["uī"])
  end

  defp decline(noun, :fourth, :singular, :accusative) do
    endings(noun, ["um"])
  end

  defp decline(noun, :fourth, :singular, :ablative) do
    endings(noun, ["ū"])
  end

  defp decline(noun, :fourth, :plural, grammatical_case) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(noun, ["ūs"])
  end

  defp decline(noun, :fourth, :plural, :genitive) do
    endings(noun, ["uum"])
  end

  defp decline(noun, :fourth, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    endings(noun, ["ibus"])
  end

  #############################################################################
  # decline/4 for fifth declension nouns.
  #############################################################################
  defp decline(noun, :fifth, :singular, :vocative) do
    all_nom(noun)
  end

  defp decline(noun, :fifth, :singular, :dative) do
    all_gen(noun)
  end

  defp decline(noun, :fifth, :singular, :accusative) do
    endings(noun, ["em"])
  end

  defp decline(noun, :fifth, :singular, :ablative) do
    endings(noun, ["ē"])
  end

  defp decline(noun, :fifth, :plural, grammatical_case) when grammatical_case in [:nominative, :accusative, :vocative] do
    endings(noun, ["ēs"])
  end

  defp decline(noun, :fifth, :plural, :genitive) do
    endings(noun, ["ērum"])
  end

  defp decline(noun, :fifth, :plural, grammatical_case) when grammatical_case in [:dative, :ablative] do
    endings(noun, ["ēbus"])
  end

end
